arXiv is now an independent nonprofit! Learn more
License: CC BY 4.0
arXiv:2609.12471v1 [cs.CL] 11 Sep 2026

AMDKernelVault: Large-Scale Datasets and Agentic Training for
AMD GPU Kernel Optimization

Ji Liu    Saptarshi Majumder    Yiqing Huang    Wenwen OuyangUmang PandeyZeping LiChushi ChenZihao AnPuyuan YangZekai LiSina RafatiZiqiong LiuPratik Prabhanjan BrahmaDong LiZicheng LiuSharon ZhouEmad BarsoumAdvanced Micro Devices, Inc.*Equal contribution.Correspondence: {liuji,sapmajum,Cody.Huang,viouyang}@amd.com
Abstract

We introduce AMDKernelVault, an open HIP and Triton kernel corpus and training framework for recent AMD CDNA GPUs. Existing LLM-based kernel agents are largely CUDA/NVIDIA-centric and often depend on repeated frontier-LLM calls for generation, reflection, and optimization. To address this gap, we develop HIPKernelGen and TritonKernelGen, agent-driven pipelines that transform PyTorch references into HIP or Triton kernels, compile and validate candidates under ROCm, and latency-profile them on AMD hardware. The corpus contains 62,153 execution-verified HIP kernel samples, 2,377 production-grounded ROCm Libraries QA entries, and 39,893 Triton kernels. We further train Qwen3-8B with supervised fine-tuning and execution-aware reinforcement learning as a demonstration of the corpus’s utility. Under fixed evaluation budgets, it achieves the highest correctness among the compared models on PyTorch-to-HIP (34.0% Pass@1), TritonBench-G (33.2% Corr@3), and ROCmBench (41.94% Corr@3), but does not uniformly lead compilation or speed metrics. The corpus and documentation are available at https://huggingface.co/datasets/amd/AIG-Datasets, and the associated training and kernel-generation code is available at https://github.com/AMD-AGI/hip_kernel_llm_lab.

1 Introduction

Recent advances in GPU kernel optimization have been driven by LLM-based (Chen et al., 2025; Baronio et al., 2026) and agent-based approaches (Wang et al., 2025a; Dai et al., 2026), together with standardized benchmarks such as KernelBench (Ouyang et al., 2025; Lange et al., 2025b; Heakl et al., 2026). These agent-based systems leverage compiler feedback, runtime correctness checks, and performance profiling to iteratively refine generated kernels (Wang et al., 2025a; Chen et al., 2025). However, the data and tooling behind these systems remain largely CUDA/NVIDIA-centric. In contrast, AMD’s HIP/ROCm (Advanced Micro Devices, Inc., 2025c; Advanced Micro Devices, Inc., 2025a) ecosystem lacks large-scale, execution-verified kernel data for training and evaluating kernel LLMs, especially across both native HIP and Triton backends.

This data gap is not a matter of syntax conversion alone. Although CUDA and HIP expose similar programming abstractions, performant AMD kernels depend on ROCm compilation behavior, wavefront execution, LDS usage, memory coalescing, matrix-core support, and Triton (Tillet et al., 2019) backend constraints. Direct translation through hipify (Advanced Micro Devices, Inc., 2025b) may preserve API structure while retaining NVIDIA-shaped tiling constants, warp-size assumptions, input-specialized indexing, or memory layouts that fail to compile or underperform on AMD GPUs. Reliable AMD kernel data therefore requires native compilation, execution-based correctness validation, and latency measurement under ROCm.

We present AMDKernelVault, an open, execution-verified HIP and Triton (Tillet et al., 2019) kernel corpus together with scalable agent-based pipelines for AMD GPU kernel data production. Our pipelines, HIPKernelGen and TritonKernelGen, start from PyTorch references and use a generate–evaluate–reflect loop to synthesize candidate kernels, compile them under ROCm, validate numerical correctness across test inputs, and profile latency on AMD hardware. This process produces 64K+ validated HIP/ROCm data and \sim40K Triton kernels for recent AMD CDNA GPUs, including benchmark-derived tasks, synthesized operator workloads, and production-grounded supervision from rocBLAS and rocSOLVER.

Beyond releasing data, we evaluate whether this corpus can train compact AMD kernel LLMs that reduce reliance on frontier LLMs in agentic optimization systems. Existing kernel-agent workflows often call frontier models repeatedly for generation, reflection, and optimization, which increases cost and limits deployment in private AMD environments. Using AMDKernelVault, we train Qwen3-8B (Yang et al., 2025) with supervised fine-tuning and execution-aware reinforcement learning, then insert the trained model into a GEAK-style (Wang et al., 2025a) agent loop where one local policy serves as generator, reflector, and optimizer. The training recipe is intended as an effectiveness test for the corpus and pipeline rather than a new RL algorithm.

Under the same evaluation budget, the trained 8B model achieves the highest correctness among the compared models on PyTorch-to-HIP (34.0% Pass@1) and TritonBench-G (33.2% Corr@3), while running locally on AMD hardware. These results show that execution-verified AMD kernel data can support locally deployable kernel LLMs for ROCm environments. Our primary contribution is the resource and its AMD-native generation and validation infrastructure; the SFT+RL recipe is an effectiveness demonstration rather than a new learning algorithm. Comparisons with frontier models are specific to the reported correctness metrics, benchmarks, and agent budgets.

Our contributions:
  • AMD-native kernel corpus. We release 64K+ HIP/ROCm and \sim40K Triton data validated through ROCm compilation, execution-based correctness tests, latency profiling, or production-grounded HIP code supervision on recent AMD CDNA GPUs.

  • Scalable data production pipelines. HIPKernelGen and TritonKernelGen provide reusable agent-based generate–evaluate–reflect workflows for producing verified HIP and Triton kernel data from PyTorch references, including benchmark-derived, synthesized, and production-grounded sources.

  • Compact AMD kernel LLMs. As a corpus-utility demonstration, we train a local Qwen3-8B model to serve the generator, reflector, and optimizer roles in a GEAK-style loop, improving specific AMD HIP/Triton correctness metrics under the same evaluation budget.

2 Related Work

GPU Kernel Optimization.

GPU kernel optimization is a longstanding problem in high-performance ML systems (Ragan-Kelley et al., 2013; Chen et al., 2018; Vasilache et al., 2018). Significant efforts have focused on operator fusion, tiling, memory hierarchy optimization, and custom kernel design. Representative systems such as FlashAttention (Dao et al., 2022; Shah et al., 2024) show that carefully designed kernels can provide substantial speedups for modern ML workloads. More recently, automated frameworks have used program analysis, search, and learning-based techniques (Zheng and others, 2020) to reduce the manual effort required for kernel tuning. However, most of these methods are developed and evaluated primarily on Nvidia GPUs and CUDA backends, leaving AMD-focused kernel optimization comparatively underexplored.

Refer to caption
Figure 1: Unified kernel generation framework. Stage 1 standardizes PyTorch modules into function-style references. Stage 2 synthesizes HIP and Triton kernels using a GEAK-based agentic loop with compilation, correctness validation, and latency-guided optimization.
HIP, Triton, and ROCm.

AMD’s HIP provides a CUDA-like programming model within the ROCm software stack (Advanced Micro Devices, Inc., 2025c; Advanced Micro Devices, Inc., 2025a), while Triton offers a Python-based DSL for writing custom GPU kernels with explicit control over memory and parallelism (Tillet et al., 2019). Although these abstractions improve portability, AMD execution differs in ROCm compiler behavior, wavefront-level execution, LDS usage, memory coalescing, and backend scheduling constraints. Source-to-source tools such as hipify can preserve syntax but provide limited guarantees of correctness or performance portability (Advanced Micro Devices, Inc., 2025b). This motivates native HIP/ROCm validation rather than treating AMD support as a post-hoc translation step.

GPU Kernel Datasets and Benchmarks.

High-quality datasets are essential for advancing kernel optimization research. KernelBench (Ouyang et al., 2025) provides standardized environments for generating and evaluating optimized kernels, while other benchmarks cover specific operator families (Che et al., 2009). However, existing resources remain largely CUDA-centric. CASS (Heakl et al., 2026) facilitates CUDA-to-HIP translation, but does not provide large-scale PyTorch-to-HIP, HIP-to-HIP, or AMD Triton data. Existing Triton benchmarks also primarily target NVIDIA assumptions that may not generalize to ROCm. AMDKernelVault addresses this gap by constructing a unified HIP/Triton corpus validated through compilation, execution-based correctness checks, and latency profiling on AMD GPUs.

LLM-Based Kernel Agents.

Recent work explores LLM-based agents for kernel synthesis and optimization (Chen et al., 2025; Baronio et al., 2026; Li et al., 2025b; Wang et al., 2025a). Kevin (Baronio et al., 2026) demonstrates multi-turn RL for CUDA kernels, AutoTriton (Li et al., 2025b) and TritonRL (Woo et al., 2025) study execution-aware training for Triton generation, and GEAK (Wang et al., 2025a) provides an agentic framework for Triton optimization. AVO (Chen et al., 2026) further explores long-horizon autonomous kernel evolution on NVIDIA GPUs. AMDKernelVault is complementary: it focuses on execution-verified AMD HIP/ROCm/Triton data and compact local policies for generate–evaluate–reflect loops. Additional discussion is provided in Appendix B.

3 Agent-Based Dataset Generation

We present a unified framework for generating execution-verified GPU kernels across both HIP and Triton, using a shared generate–evaluate–reflect paradigm (Figure 1).

3.1 Unified Pipeline Architecture

We propose a unified, agent-driven framework for large-scale GPU kernel data construction. The framework follows a shared generate–evaluate–reflect paradigm, enabling systematic synthesis, validation, and refinement of GPU kernels grounded in standardized PyTorch references. Within this unified design, we instantiate two concrete pipelines—HIPKernelGen and TritonKernelGen—which specialize the same agent-driven workflow to the HIP and Triton ecosystems respectively.

Stage 1: PyTorch Reference Standardization.

For each target operator or module kernel, we construct a function-style PyTorch reference derived from the PyTorch module input. This representation ensures reliable invocation, consistent interface semantics, and execution-based validation across both HIP and Triton backends. When available, compiler-generated kernels (e.g., Torch Inductor Triton code) are used for cross-checking but do not replace reference-based correctness.

Stage 2: Backend-Specific Kernel Synthesis.

Given the standardized PyTorch reference, backend-specific agents generate candidate GPU kernels. Each candidate is subjected to compilation, execution-based correctness checks against the PyTorch reference, and latency profiling. The default HIP checker uses torch.allclose with rtol=1e-4, atol=1e-4, and equal_nan=True. For Triton, task-specific tolerances may be configured from the reference task; otherwise, the fallback is rtol=1e-4, atol=1e-3, and equal_nan=False. We use no fixed dtype- or operator-specific rule. Failures trigger reflector-guided regeneration until correctness criteria are met or the attempt limit is reached.

Test-case construction and validation.

For PyTorch\rightarrowHIP and PyTorch\rightarrowTriton, test inputs are inherited from CUDA-Agent-Ops-6K (Dai et al., 2026), GPUMode (Paliskara and Saroufim, 2025) and AI-CUDA-Engineer-Archive (Lange et al., 2025a) when available and augmented with verified LLM-generated cases to cover additional shapes, strides, dtypes, and boundary conditions. HIP kernels, including HIP\rightarrowHIP optimization targets, are bound through torch.utils.cpp_extension so that generated kernels can be executed under the same PyTorch reference tests. For Triton\rightarrowTriton, we extend TritonBench-style tests with ROCm-compatible cases and discard candidates that only pass input-specialized or shape-specialized checks. Thus, LLM-generated tests are not accepted as supervision by themselves; they are used as additional probes inside an execution-based validation loop.

HIPKernelGen.

HIPKernelGen instantiates the unified framework for HIP kernels targeting AMD GPUs. Much of the HIP corpus was generated from August to December 2025 using GPT-5, a strong available kernel-generation model during collection. Each candidate is compiled via hipcc, validated against PyTorch references, and profiled for latency and speedup. Retention depends on ROCm execution rather than generator identity.

TritonKernelGen.

TritonKernelGen mirrors the HIP pipeline for Triton kernels. We use GPT-oss-120B (Agarwal et al., 2025), DeepSeek-R1 (Guo et al., 2025), and Qwen2.5-32B (Qwen et al., 2025) to diversify implementations and reasoning traces. Each kernel is compiled and executed via Triton 3.3.0 and tested on approximately ten correctness and three performance configurations spanning shapes, strides, FP32/FP16/BF16 where applicable, and boundary cases.

3.2 Source Data

Our kernel datasets are built upon six complementary sources spanning synthesized CUDA-agent tasks, GitHub-derived PyTorch programs, curated Triton corpora, benchmark-style kernels, and production ROCm libraries.

3.2.1 CUDA-Agent-Ops-6K Dataset

The CUDA-Agent-Ops-6K dataset (Dai et al., 2026) contains 6,000 synthesized operator-level PyTorch tasks for CUDA kernel generation and optimization. It is built from torch/transformers seed operators, LLM-composed multi-operator tasks, and execution-based filtering. We use it as a source for HIP generation, but the original dataset remains CUDA-centric and does not directly support HIP or ROCm.

3.2.2 GPUMODE-KernelBook Dataset

The GPUMODE-KernelBook (Paliskara and Saroufim, 2025) dataset contains approximately 18,000 PyTorch modules paired with Torch Inductor Triton implementations from GitHub-derived workloads. It provides broad operator coverage across primitives, fused kernels, and larger model components, but remains CUDA/Triton-centric and does not directly provide AMD-validated HIP kernels.

3.2.3 AI-CUDA-Engineer-Archive Dataset

The AI-CUDA-Engineer-Archive (Lange et al., 2025a) dataset is built with the AI CUDA Engineer framework from over 250 KernelBench PyTorch benchmarks. It provides PyTorch references, generated CUDA kernels, diagnostics, and profiling results, but remains CUDA-centric and does not directly support HIP.

3.2.4 Stack-v2-dedup-Triton Dataset

We incorporate the curated Stack-v2-dedup-Triton (Lozhkov et al., 2024) dataset, constructed by filtering approximately 13,000 candidate samples to obtain 2,269 high-quality Triton–PyTorch pairs. This source provides medium-difficulty Triton examples emphasizing correctness and moderate complexity.

3.2.5 TritonBench-8k Dataset

The TritonBench-8k (Li et al., 2025a) dataset consists of samples derived from web crawl data and synthetic code generated via the Ninetoothed DSL (Huang et al., 2025). We apply targeted curation to remove DSL-specific artifacts and ensure correctness across diverse shapes, strides, and precision types.

3.2.6 ROCm Libraries Dataset Source

As a production-oriented source, we leverage AMD’s ROCm Libraries monorepo, focusing on rocBLAS and rocSOLVER. Unlike benchmark-style datasets, these libraries capture engineering-tested HIP kernels, launch-side logic, and idiomatic production patterns.

3.3 Generated Kernel Corpus

Table 1 summarizes our released corpus. HIP and Triton kernel entries are processed through HIPKernelGen/TritonKernelGen with execution-based validation on AMD GPUs. The ROCm Libraries QA subset provides complementary production-grounded supervision rather than generated kernel-pair data.

Dataset Samples
HIP/ROCm Datasets
HIP-CudaAgent 5,388
HIP-GPUMode 22,397
HIP2HIP (Optimization) 34,368
ROCm Libraries QA 2,377
Total HIP/ROCm 64,530
Triton Datasets
Triton-Stack 2,269
Triton-Bench (from TritonBench-8k) 7,713
Triton-GPUMode 18,000
Triton-AICE 11,911
Total Triton 39,893
Table 1: Summary of the corpus used in this paper. HIP and Triton kernel samples are execution-validated on AMD GPUs; ROCm Libraries QA provides production-grounded supervision.
Counting and audit.

The 64,530 HIP/ROCm total comprises 62,153 execution-verified HIP kernel samples and 2,377 ROCm Libraries QA entries. It excludes a separately released 14,282-entry HIP-AICUDA subset that was not used in this paper. HIP-CudaAgent retains 5,388 of 6,000 upstream tasks after excluding 612 incomplete records; HIP-GPUMode contains 5,910 unique tasks and 22,397 kernel variants; and HIP2HIP contains 31,155 single-turn and 3,213 multi-turn records. Triton counts use 39,893 unique source-level samples; packaged reasoning variants may yield a different row count. Since rejection logs were not aggregated uniformly across source pipelines, we report retained counts and subset-specific validation rather than a single corpus-wide failure rate.

HIP Datasets.

HIP-CudaAgent: Built from CUDA-Agent-Ops-6K using HIPKernelGen. We adapt the synthesized PyTorch operator tasks into function-style references and use the generate–evaluate–reflect pipeline to produce HIP implementations. Each candidate is compiled with the ROCm toolchain, checked against the PyTorch reference, and filtered by execution-based correctness validation. This yields 5,388 validated PyTorch-to-HIP samples.

HIP-GPUMode: Generated from GPUMODE-KernelBook using HIPKernelGen. We first convert PyTorch modules to function-style references, then generate HIP kernels using frontier LLMs. Each candidate undergoes compilation, execution-based validation, and latency profiling. The generate–evaluate–reflect loop iterates until correctness is achieved or a maximum attempt limit is reached. This produces 5,910 unique PyTorch entries with 22,397 HIP kernel variants (avg 3.8 variants/entry).

HIP2HIP: We extend HIPKernelGen to kernel-to-kernel optimization by operating directly on existing HIP kernel functions or operators. Using HIP-GPUMode as a base, we generate 34,368 validated HIP kernel optimization pairs.

ROCm Libraries QA.

The ROCm Libraries QA subset contains production-grounded interface-level supervision rather than generated kernel-pair data. For each interface in rocBLAS and rocSOLVER, we identify kernel-relevant code regions and generate QA pairs targeting computation semantics, index-to-tensor mappings, memory access patterns, synchronization strategies, and edge cases. This yields 1,858 entries from rocBLAS and 519 from rocSOLVER (2,377 total), capturing production HIP development patterns.

Triton Datasets.

TritonKernelGen processes all Triton sources through a unified validation pipeline, validating numerical correctness across diverse tensor shapes, strides, and dtypes (FP32/FP16/BF16) on AMD GPUs.

Triton-Stack: Curated from Stack-v2-dedup, filtering \sim13K candidates to 2,269 medium-difficulty Triton–PyTorch pairs with verified AMD compatibility, then generated and execution-verified with TritonKernelGen.

Triton-Bench: Derived from TritonBench-8k, removing DSL-specific artifacts and validating 7,713 samples, generated and execution-verified with TritonKernelGen.

Triton-GPUMode: Sourced from GPUMODE-KernelBook (\sim18K entries), with Triton implementations ported from torch-compile kernels to generic kernels able to handle all shapes and params, generated and execution-verified with TritonKernelGen.

Triton-AICE: From AI-CUDA-Engineer-Archive (11,911 samples), converting CUDA-centric Triton to AMD-validated implementations with TritonKernelGen.

Together, this yields 62,153 validated HIP kernel samples and 39,893 validated Triton–PyTorch samples for AMD GPUs, plus 2,377 production-grounded ROCm Libraries QA entries. Additional details on dataset construction, source-specific processing, metadata, and difficulty distributions are provided in Appendix C and Appendix D.

3.4 Production-Grounded ROCm QA Supervision

To extend GPU kernel data construction beyond benchmark-style kernels, we extract HIP-kernel-focused supervision from production-grade ROCm libraries at the interface level. This subset is not treated as direct execution-verified kernel-pair data; instead, it provides production-grounded QA supervision anchored to retrieved rocBLAS and rocSOLVER code.

Interface decomposition.

For each target project, we start from a user-facing interface and decompose it into a compact kernel-relevant scope. We traverse the call graph from the interface entrypoint to identify code regions directly involved in GPU execution, including device-side kernels, host-side launch logic, parameter preparation, and tightly coupled helper utilities. This interface-level view preserves the connection between public APIs and the underlying HIP kernels, which is often lost when treating files or repositories as unstructured text.

Scope control.

Production libraries contain substantial infrastructure code unrelated to kernel behavior. To avoid recursion blow-up and irrelevant context, we use project-specific allowlists and disallowlists. The disallowlist filters logging, debugging, testing, and auxiliary utilities, while the allowlist constrains traversal to kernel-relevant subsystems. Because rocBLAS and rocSOLVER differ in organization and dispatch structure, each project uses tailored traversal rules.

Kernel-centric QA generation.

Given the retrieved interface context, an agent constructs HIP-kernel-focused QA pairs. The questions target concrete implementation aspects, including computation semantics, index-to-tensor mappings, memory access patterns, synchronization and reduction strategies, launch configuration, and correctness constraints. Answers are grounded in the retrieved code context rather than generic library descriptions, so the resulting supervision emphasizes actionable kernel logic.

Applying this pipeline yields 519 interface-level entries from rocSOLVER, comprising 226 kernel-implementation-focused samples and 293 kernel-centric QA explanations. From rocBLAS, we extract 1,858 entries, including 1,169 kernel-implementation samples and 689 QA-style explanations. Together, these 2,377 production-grounded entries complement benchmark-derived HIP and Triton datasets with real-world host–device orchestration and production HIP development patterns.

3.5 Difficulty Distribution

We classify samples into three difficulty levels for curriculum training: Level 1 covers standalone single-function kernels, Level 2 covers fused operators or single-file implementations, and Level 3 covers multi-file kernels with cross-module dependencies. The full Triton corpus comprises 11,824 L1 (29.6%), 24,765 L2 (62.1%), and 3,304 L3 (8.3%) samples; the curriculum-eligible subset used per Triton RL phase is reported in Appendix G. The ROCm Libraries QA data follows a similar distribution: rocSOLVER (190 L1, 198 L2, 131 L3) and rocBLAS (971 L1, 506 L2, 381 L3).

4 Training Compact AMD Kernel LLMs

AMDKernelVault is designed not only as a kernel corpus, but also as a training resource for compact AMD kernel LLMs. We use it to train Qwen3-8B with a two-stage recipe: supervised fine-tuning (SFT) for cold-start kernel generation, followed by execution-aware reinforcement learning (RL) that uses compilation, correctness, and latency feedback.

4.1 Kernel Tasks and Metrics

Our datasets support four kernel-centric tasks: PyTorch\rightarrowHIP translates PyTorch modules into semantically equivalent HIP implementations; HIP\rightarrowHIP optimizes existing HIP kernels; Text\rightarrowTriton generates Triton kernels from natural-language specifications; and Triton\rightarrowTriton improves Triton kernels through tiling, memory-access, and boundary-handling changes.

We evaluate kernels with three execution-based metrics: compilation pass, successful build under the target toolchain; correctness pass, numerical equivalence against the PyTorch reference; and speed, kernel-level speedup over the input kernel or PyTorch baseline. Comp@kk and Corr@kk report compilation and correctness after kk agent iterations, while Pass@kk measures the probability of obtaining at least one correct result from kk samples. For PyTorch-to-Triton, speedup is measured against the original PyTorch implementation; for Triton-to-Triton, it is measured against the input kernel. Speed@10 averages the speedups of the best ten correct samples.

4.2 Supervised Fine-Tuning

Direct RL from a pretrained model is difficult in this domain because early rollouts often fail to compile or violate kernel interfaces, producing sparse feedback. We therefore first apply SFT to teach kernel syntax, binding conventions, memory-access idioms, and common implementation patterns.

For HIP, we use instruction–response pairs from the HIP portions of AMDKernelVault, where the input is a PyTorch task or source HIP kernel and the output is a validated HIP implementation. For Triton, we use reasoning-augmented examples when available, so that the model learns to plan tiling, masking, boundary checks, and memory layouts before emitting code. This stage provides the cold-start capability needed for subsequent execution-aware RL.

4.3 Execution-Aware Reinforcement Learning

After SFT, we optimize the model using execution feedback. HIP training uses single-turn GRPO, where each candidate is compiled, tested, and scored after one generation attempt. Triton training uses a multi-turn variant aligned with the GEAK-style generate–evaluate–reflect loop: the model proposes a kernel, observes compiler or runtime feedback, reflects on the failure or performance bottleneck, and generates an improved candidate.

The reward combines three signals: compilation, correctness, and speedup. Compilation credit provides partial feedback for syntactically and interface-compatible kernels, correctness rewards functional equivalence, and a bounded speedup term encourages performance improvements without allowing timing outliers to dominate learning. This reward is designed for the low-success regime of AMD kernel generation, where binary correctness rewards alone provide too little gradient signal. Full reward definitions and GRPO details are provided in Appendix E.

4.4 Agentic Integration

A key goal is to test whether the corpus can train a compact local model for an agentic kernel optimization loop. We train one Qwen3-8B policy to serve the generator, reflector, and optimizer roles. During multi-turn Triton RL, the same model produces a candidate, interprets execution feedback, and emits the next revision.

This training recipe is not presented as a new RL algorithm. Rather, it evaluates whether execution-verified AMDKernelVault data can specialize a compact policy for AMD kernel agents. Section 5 therefore reports controlled, metric-specific comparisons under identical agent budgets.

4.5 Training Configuration

We train Qwen3-8B with SFT followed by RL on AMD MI-series GPUs. HIP training uses SFT followed by single-turn GRPO, while Triton training uses SFT followed by multi-turn on-policy reflection RL with curriculum scheduling. We keep the main text focused on the training recipe used to evaluate AMDKernelVault as a corpus for compact AMD kernel LLMs. Appendix E provides task serialization, SFT and GRPO objectives, reward implementation, curriculum schedule, hyperparameters, and additional analysis of reward sensitivity.

5 Experiments

We evaluate whether AMDKernelVault can train compact AMD kernel LLMs that improve HIP and Triton kernel generation under execution-based validation. We compare Qwen3-8B trained on our corpus against frontier LLMs (GPT-5, Gemini 2.5 Pro, Claude Sonnet 4). For HIP, we evaluate PyTorch-to-HIP translation on 200 KernelBench Level-1/2 tasks (Ouyang et al., 2025) and HIP-to-HIP optimization on 100 held-out samples. For Triton, we use TritonBench-G (184 kernels) and ROCmBench (31 AMD-specific kernels) with the same 3-iteration OptimAgent-v2 protocol (Wang et al., 2025b). Each model serves as generator, reflector, and optimizer under identical agent settings, and all reported Triton inference runs use MI355X GPUs.

Train/test separation and overlap controls.

Evaluation cases are excluded from training by source and task identifier. HIP evaluation uses 200 held-out KernelBench-derived translation tasks and 100 held-out HIP-to-HIP optimization tasks; Triton evaluation uses the curated 184-kernel TritonBench-G and 31-kernel ROCmBench sets. These controls prevent exact task-identifier reuse. However, because generated kernels can implement semantically similar operator patterns using different code, exact-match filtering cannot exclude every functional near-duplicate. We did not conduct an exhaustive corpus-wide semantic-overlap search and therefore interpret the reported results as benchmark-specific rather than as evidence of overlap-free generalization.

Evaluation protocol and metrics.

All Triton model comparisons use the same OptimAgent-v2 configuration: num_offsprings=1, max_iteration=3, identical generator/reflector/optimizer role assignment, and no profiler feedback. Thus, each row changes the model checkpoint while holding the external agent scaffold and execution budget fixed. Comp@kk and Corr@kk report the fractions of tasks that compile and pass numerical validation by iteration kk, respectively; Pass@kk reports the probability of obtaining at least one correct result from kk independently sampled candidates. For PyTorch-to-kernel generation, speedup is measured against the PyTorch reference; for kernel-to-kernel optimization, it is measured against the input kernel. Speed comparisons include only correctness-passing candidates.

Benchmark coverage.

The evaluation covers attention and softmax variants, layer and RMS normalization, GEMM-family operations, elementwise transformations, reductions, fused blocks, MoE-style routing, and online-softmax-style fused reductions. KernelBench Level-1/2 tests emphasize operator implementation and composition, while TritonBench-G spans difficulty levels from standalone operators to harder fused kernels; ROCmBench adds AMD-specific cases. This breadth tests multiple common deep-learning kernel patterns, but it is not a per-family performance study: the benchmark sizes are limited, and we do not report operator-family-specific pass rates.

Triton reproducibility setup.

Triton workflows use Triton 3.3.0. Corpus validation and profiling use MI325 and MI350 GPUs; Triton SFT and RL use 32 MI325X GPUs; and reported Triton inference uses MI355X GPUs. The corpus targets gfx942 for MI325X stages and gfx950 for MI350/MI355X stages. ROCm, PyTorch, and Python versions vary across data-production snapshots and are not pinned to one release. Warm-up and measurement use 25 ms and 100 ms time budgets, respectively, with per-kernel iteration counts derived from a short timing estimate. We report median GPU-event timing, flush L2 before each measured call, synchronize after the measured loop, and discard no timing observations.

5.1 HIP Kernel Results

Table 2 evaluates PyTorch-to-HIP translation. Frontier LLMs compile many candidates but achieve limited correctness; GPT-5 reaches 75.5% compilation but only 24.0% correctness. Training on AMDKernelVault substantially improves Qwen3-8B: SFT raises correctness to 31.5%, and RL further improves it to 34.0%, outperforming all frontier baselines in correctness.

Model Comp Acc Corr Acc
Gemini 2.5 Pro 70.5% 18.5%
Claude Sonnet 4 85.0% 22.0%
GPT-5 75.5% 24.0%
Q3-8B (base) 50.0% 14.5%
Q3-8B + SFT 78.0% 31.5%
Q3-8B + SFT + RL 83.5% 34.0%
Table 2: Pass@1 results of PyTorch-to-HIP translation (200 tasks from KernelBench Levels 1 and 2). Q3-8B denotes Qwen3-8B.

Table 3 evaluates HIP-to-HIP optimization. The trained model reaches 84% correctness, close to GPT-5 (87%) and substantially above the base model (18%), while maintaining positive speedup. These results indicate that the HIP portion of AMDKernelVault supports both translation and optimization tasks.

Model Comp Acc Corr Acc Speed@10
Gemini 2.5 Pro 81% 69% 1.15×\times
Claude Sonnet 4 93% 67% 1.11×\times
GPT-5 89% 87% 1.32×\times
Q3-8B (base) 26% 18% 1.02×\times
Q3-8B + SFT 86% 70% 1.06×\times
Q3-8B + SFT + RL 90% 84% 1.14×\times
Table 3: Pass@1 results of HIP-to-HIP kernel optimization (100 held-out modules). Q3-8B denotes Qwen3-8B and speed@10 denotes average speedup achieved over baseline of best 10 correct samples.

5.2 Triton Kernel Results

Table 4 reports TritonBench-G results under the fixed 3-iteration GEAK OptimAgent-v2 evaluator. The base Qwen3-8B model reaches 4.9% Corr@3, while SFT improves to 10.3% and SFT+RL reaches 33.2%, compared with GPT-5 at 15.2% and Claude Sonnet 4 at 29.8%. Since all rows use the same scaffold and budget, this controlled comparison demonstrates the corpus’s utility for specializing a compact local model.

Model Corr@1 Corr@3 Comp@3 Speed
Gemini 2.5 Pro 8.7% 23.4% 32.6% 1.31×\times
Claude Sonnet 4 7.6% 29.8% 46.7% 1.34×\times
GPT-5 3.8% 15.2% 28.8% 1.05×\times
Q3-8B (base) 0.0% 4.9% 14.7% 0.97×\times
Q3-8B + SFT 4.3% 10.3% 27.7% 1.19×\times
Q3-8B + SFT + RL 12.0% 33.2% 67.9% 1.46×\times
Table 4: TritonBench-G results under the fixed 3-iteration GEAK OptimAgent-v2 evaluator. All models use the same agent budget and role assignment; only the LLM checkpoint changes. Corr/Comp denote correctness/compile accuracy, and Q3-8B denotes Qwen3-8B.

Table 5 shows results on ROCmBench. The trained Qwen3-8B model obtains the highest Corr@3 (41.94%), compared with Claude Sonnet 4 (35.48%) and GPT-5 (29.03%). Claude Sonnet 4 remains stronger on Comp@3 (67.74% versus 58.06%) and speed (1.82×\times versus 1.61×\times), so the advantage is specific to correctness.

Difficulty Analysis.

Our model leads the compared frontier baselines on D1–D3 of TritonBench-G, but Claude Sonnet 4 solves more D4 kernels (18/84 versus 13/84), and D5 remains unsolved by all models. Full results are provided in Appendix K.

Model Corr@1 Corr@3 Comp@3 Speed
Gemini 2.5 Pro 12.90% 25.81% 41.94% 1.47×\times
Claude Sonnet 4 19.35% 35.48% 67.74% 1.82×\times
GPT-5 16.13% 29.03% 35.48% 1.31×\times
Q3-8B (base) 3.22% 6.45% 19.35% 0.74×\times
Q3-8B + SFT + RL 22.58% 41.94% 58.06% 1.61×\times
Table 5: ROCmBench (AMD) results under the fixed 3-iteration GEAK OptimAgent-v2 evaluator. All rows use the same agent budget, and Q3-8B denotes Qwen3-8B.
Limitations and Failure Analysis.

Despite strong results on D1–D3 tasks, expert-level kernels remain challenging. Common failure modes include incorrect tile/block size choices for unusual tensor shapes, synchronization errors in multi-stage reductions, and memory-coalescing failures under AMD’s wavefront execution model. These failures suggest that future work should incorporate richer profiler-guided feedback and broader expert-kernel supervision.

5.3 Ablation Studies

We conduct ablations to assess the importance of data scale, SFT cold start, and multi-turn RL. Performance improves with more training data, SFT provides a critical initialization before RL, and multi-turn RL contributes +13.1% on Triton. Detailed ablation tables and reward-sensitivity results are provided in Appendix J.

6 Conclusion

AMDKernelVault is an open HIP and Triton kernel resource for recent AMD CDNA GPUs, supported by AMD-native generation and validation infrastructure. As a utility demonstration, we specialize Qwen3-8B for a GEAK-style loop; under fixed budgets, it achieves the highest correctness on selected benchmarks but does not uniformly lead compilation or speed. We release the corpus, documentation, and code to support future AMD kernel research.

Limitations

Our evaluation is limited to AMD GPU environments and kernel-generation benchmarks available to us. Expert-level kernels remain difficult across all tested models, and the reported performance depends on the specific ROCm, Triton, and hardware configurations used during validation. Although we include production-derived ROCm library supervision, some generated kernels may still inherit suboptimal structures from upstream CUDA/Triton sources or from LLM-generated candidates.

Several gaps remain for future work: broader corpus-level profiling of latency, memory footprint, register pressure, and occupancy; per-operator-family pass-rate analysis beyond the current difficulty breakdown; larger held-out HIP evaluation sets; and evaluation beyond deep-learning workloads, including HPC, sparse/graph, and database kernels. Future work should also incorporate richer profiler-guided feedback and broader expert-kernel supervision.

We do not report a corpus-wide semantic near-duplicate audit or uniformly aggregated candidate-rejection rates because source pipelines use different logging formats. Software environments also vary across data-production snapshots, although Triton is fixed to version 3.3.0 for the reported Triton workflow.

Ethical Considerations and Impact Statement

This work focuses on GPU kernel generation and optimization for AMD hardware. The released artifacts consist of program code, generated kernels, execution metadata, and QA supervision grounded in public ROCm library code. Provenance, usage conditions, and per-source license notes are included with the public artifact.

The expected positive impact is improved efficiency for AI workloads. Faster generated kernels can reduce training and inference runtime, while a compact local model can reduce repeated frontier-LLM calls during kernel search. These improvements may lower computational cost and energy use for AMD GPU deployments.

As with other code-generation systems, generated code may be incorrect, inefficient, or unsafe if deployed without validation. We mitigate this risk by emphasizing execution-based compilation, correctness testing, and latency profiling throughout data construction and evaluation. The work is narrowly focused on performance engineering for GPU kernels, and generated kernels should be validated in the target environment before production use.

References

  • Advanced Micro Devices, Inc. (2025a) Advanced Micro Devices, Inc. HIP documentation (programming guide). Note: https://rocm.docs.amd.com/projects/HIP/en/latest/Accessed 2025-12-15 Cited by: §1, §2.
  • Advanced Micro Devices, Inc. (2025b) Advanced Micro Devices, Inc. HIPIFY documentation. Note: https://rocm.docs.amd.com/projects/HIPIFY/en/latest/Accessed 2025-12-15 Cited by: §1, §2.
  • Advanced Micro Devices, Inc. (2025c) Advanced Micro Devices, Inc. ROCm documentation. Note: https://rocm.docs.amd.com/Accessed 2025-12-15 Cited by: §1, §2.
  • Agarwal et al. (2025) S. Agarwal, L. Ahmad, J. Ai, S. Altman, A. Applebaum, E. Arbus, R. K. Arora, Y. Bai, B. Baker, H. Bao, et al. Gpt-oss-120b & gpt-oss-20b model card. arXiv preprint arXiv:2508.10925. Cited by: §3.1.
  • Baronio et al. (2026) C. Baronio, P. Marsella, B. Pan, S. Guo, and S. Alberti Kevin: multi-turn rl for generating cuda kernels. In International Conference on Learning Representations, Cited by: Appendix B, Table 6, §1, §2.
  • Che et al. (2009) S. Che, M. Boyer, J. Meng, D. Tarjan, J. W. Sheaffer, S. Lee, and K. Skadron Rodinia: a benchmark suite for heterogeneous computing. In IEEE International Symposium on Workload Characterization (IISWC), pp. 44–54. External Links: Document Cited by: §2.
  • Chen et al. (2026) T. Chen, Z. Ye, B. Xu, Z. Ye, T. Liu, A. Hassani, T. Chen, A. Kerr, H. Wu, Y. Xu, Y. Chen, H. Chen, A. Kane, R. Krashinsky, M. Liu, V. Grover, L. Ceze, R. Bringmann, J. Tran, W. Liu, F. Xie, M. Lightstone, and H. Shi AVO: agentic variation operators for autonomous evolutionary search. External Links: 2603.24517, Link Cited by: §2.
  • Chen et al. (2018) T. Chen, T. Moreau, Z. Jiang, L. Zheng, E. Yan, H. Shen, M. Cowan, L. Wang, Y. Hu, L. Ceze, C. Guestrin, and A. Krishnamurthy TVM: an automated end-to-end optimizing compiler for deep learning. In 13th USENIX Symposium on Operating Systems Design and Implementation (OSDI 18), pp. 578–594. Cited by: §2.
  • Chen et al. (2025) W. Chen, J. Zhu, Q. Fan, Y. Ma, and A. Zou CUDA-llm: llms can write efficient cuda kernels. arXiv preprint arXiv:2506.09092. External Links: 2506.09092 Cited by: §1, §2.
  • Dai et al. (2026) W. Dai, H. Wu, Q. Yu, H. Gao, J. Li, C. Jiang, W. Lou, Y. Song, H. Yu, J. Chen, et al. Cuda agent: large-scale agentic rl for high-performance cuda kernel generation. arXiv preprint arXiv:2602.24286. Cited by: §1, §3.1, §3.2.1.
  • Dao et al. (2022) T. Dao, D. Fu, S. Ermon, A. Rudra, and C. Ré Flashattention: fast and memory-efficient exact attention with io-awareness. Advances in neural information processing systems 35, pp. 16344–16359. Cited by: §2.
  • Guo et al. (2025) D. Guo, D. Yang, H. Zhang, J. Song, P. Wang, Q. Zhu, R. Xu, R. Zhang, S. Ma, X. Bi, et al. Deepseek-r1: incentivizing reasoning capability in llms via reinforcement learning. arXiv preprint arXiv:2501.12948. Cited by: §3.1.
  • Heakl et al. (2026) A. Heakl, G. B. Stahl, S. Hashmi, S. H. E. Han, M. Ranjan, A. Kharlamova, S. Khan, and A. Mahmoud Cass: nvidia to amd transpilation with data, models, and benchmark. In Proceedings of the 64th Annual Meeting of the Association for Computational Linguistics (Volume 1: Long Papers), pp. 34489–34508. Cited by: §1, §2.
  • Huang et al. (2025) J. Huang, Z. Li, Y. Li, and H. Wang NineToothed: a triton-based high-level domain-specific language for machine learning. External Links: 2507.11978, Link Cited by: §3.2.5.
  • Lange et al. (2025a) R. T. Lange, A. Prasad, Q. Sun, M. Faldor, Y. Tang, and D. Ha The AI CUDA Engineer Archive: agentic CUDA kernel discovery, optimization and composition. Note: Hugging Face dataset, https://huggingface.co/datasets/SakanaAI/AI-CUDA-Engineer-ArchiveAccessed: 2026-08-30 Cited by: §3.1, §3.2.3.
  • Lange et al. (2025b) R. T. Lange, Q. Sun, A. Prasad, M. Faldor, Y. Tang, and D. Ha Towards robust agentic cuda kernel benchmarking, verification, and optimization. External Links: 2509.14279 Cited by: §1.
  • Li et al. (2025a) J. Li, S. Li, Z. Gao, Q. Shi, Y. Li, Z. Wang, J. Huang, H. Wang, J. Wang, X. Han, Z. Liu, and M. Sun TritonBench: benchmarking large language model capabilities for generating triton operators. In Findings of the Association for Computational Linguistics: ACL 2025, Vienna, Austria, pp. 23053–23066. External Links: Document Cited by: §3.2.5.
  • Li et al. (2025b) S. Li, Z. Wang, Y. He, Y. Li, Q. Shi, J. Li, Y. Hu, W. Che, X. Han, Z. Liu, and M. Sun AutoTriton: automatic triton programming with reinforcement learning in llms. arXiv preprint arXiv:2507.05687. External Links: 2507.05687 Cited by: Table 6, §2.
  • Lozhkov et al. (2024) A. Lozhkov, R. Li, L. B. Allal, F. Cassano, J. Lamy-Poirier, N. Tazi, A. Tang, D. Pykhtar, J. Liu, Y. Wei, et al. Starcoder 2 and the stack v2: the next generation. arXiv preprint arXiv:2402.19173. Cited by: §3.2.4.
  • Ouyang et al. (2025) A. Ouyang, S. Guo, S. Arora, A. L. Zhang, W. Hu, C. Re, and A. Mirhoseini KernelBench: can LLMs write efficient GPU kernels?. In Proceedings of the 42nd International Conference on Machine Learning, A. Singh, M. Fazel, D. Hsu, S. Lacoste-Julien, F. Berkenkamp, T. Maharaj, K. Wagstaff, and J. Zhu (Eds.), Proceedings of Machine Learning Research, Vol. 267, pp. 47356–47415. Cited by: §1, §2, §5.
  • Paliskara and Saroufim (2025) KernelBook External Links: Link Cited by: §3.1, §3.2.2.
  • Qwen et al. (2025) Qwen, A. Yang, B. Yang, B. Zhang, B. Hui, B. Zheng, B. Yu, C. Li, D. Liu, F. Huang, H. Wei, H. Lin, J. Yang, J. Tu, J. Zhang, J. Yang, J. Yang, J. Zhou, J. Lin, K. Dang, K. Lu, K. Bao, K. Yang, L. Yu, M. Li, M. Xue, P. Zhang, Q. Zhu, R. Men, R. Lin, T. Li, T. Tang, T. Xia, X. Ren, X. Ren, Y. Fan, Y. Su, Y. Zhang, Y. Wan, Y. Liu, Z. Cui, Z. Zhang, and Z. Qiu Qwen2.5 technical report. External Links: 2412.15115, Link Cited by: §3.1.
  • Ragan-Kelley et al. (2013) J. Ragan-Kelley, C. Barnes, A. Adams, S. Paris, F. Durand, and S. Amarasinghe Halide: a language and compiler for optimizing parallelism, locality, and recomputation in image processing pipelines. In Proceedings of the 34th ACM SIGPLAN Conference on Programming Language Design and Implementation (PLDI), pp. 519–530. External Links: Document Cited by: §2.
  • Shah et al. (2024) J. Shah, G. Bikshandi, Y. Zhang, V. Thakkar, P. Ramani, and T. Dao Flashattention-3: fast and accurate attention with asynchrony and low-precision. Advances in Neural Information Processing Systems 37, pp. 68658–68685. Cited by: §2.
  • Shao et al. (2024) Z. Shao, P. Wang, Q. Zhu, R. Xu, J. Song, X. Bi, H. Zhang, M. Zhang, Y. K. Li, Y. Wu, and D. Guo DeepSeekMath: pushing the limits of mathematical reasoning in open language models. External Links: 2402.03300, Link Cited by: §E.3.
  • Tillet et al. (2019) P. Tillet, H. T. Kung, and D. Cox Triton: an intermediate language and compiler for tiled neural network computations. In Proceedings of the 3rd ACM SIGPLAN International Workshop on Machine Learning and Programming Languages (MAPL ’19), pp. 10–19. External Links: Document Cited by: §1, §1, §2.
  • Vasilache et al. (2018) N. Vasilache, O. Zinenko, T. Theodoridis, P. Goyal, Z. DeVito, W. S. Moses, S. Verdoolaege, A. Adams, and A. Cohen Tensor comprehensions: framework-agnostic high-performance machine learning abstractions. CoRR. External Links: 1802.04730 Cited by: §2.
  • Wang et al. (2025a) J. Wang, V. Joshi, S. Majumder, X. Chao, B. Ding, Z. Liu, P. P. Brahma, D. Li, Z. Liu, and E. Barsoum Geak: introducing triton kernel ai agent & evaluation benchmarks. arXiv preprint arXiv:2507.23194. Cited by: Table 6, §1, §1, §2.
  • Wang et al. (2025b) J. Wang, S. Majumder, U. Pandey, C. Xu, B. Ding, M. Rao, Z. Liu, P. P. Brahma, D. Li, Z. Liu, and E. Barsoum GEAK-Triton v2 Family of AI Agents: Kernel Optimization for AMD Instinct GPUs. Note: AMD ROCm BlogsAccessed: 2026-01-27 External Links: Link Cited by: §5.
  • Woo et al. (2025) J. Woo, S. Zhu, A. Nie, Z. Jia, Y. Wang, and Y. Park TritonRL: training LLMs to think and code triton without cheating. External Links: 2510.17891, Link Cited by: §2.
  • Yang et al. (2025) A. Yang, A. Li, B. Yang, B. Zhang, B. Hui, B. Zheng, B. Yu, C. Gao, C. Huang, C. Lv, et al. Qwen3 technical report. arXiv preprint arXiv:2505.09388. Cited by: §1.
  • Zheng et al. (2020) L. Zheng et al. Ansor: generating high-performance tensor programs for deep learning. In 14th USENIX Symposium on Operating Systems Design and Implementation (OSDI 20), Cited by: §2.

Appendix A Artifact, License, and AI-Assistance Statement

Artifacts and licenses.

The data and documentation are available at https://huggingface.co/datasets/amd/AIG-Datasets. The associated training, RL, and HIP kernel-generation code is available at https://github.com/AMD-AGI/hip_kernel_llm_lab. The release includes corpus-audit, validation-protocol, and license/provenance documentation. Upstream resources remain subject to their source-specific terms: CUDA-Agent-Ops-6K and AI-CUDA-Engineer-Archive are released under CC-BY-4.0; the current GPUMODE/KernelBook release uses the June 9 Researcher Reciprocity License; TritonBench-8k uses Apache-2.0; and Stack v2 retains the BigCode and original-repository terms. ROCm Libraries are licensed per component; the source components used here include rocBLAS under the MIT License and rocSOLVER under the BSD-2-Clause License. The ROCm Libraries QA subset contains interface-level supervision and provenance references rather than a redistribution of the complete source trees. Exact source revisions and applicable terms are documented in the released provenance metadata.

Data content and filtering.

The released data consists of program code, kernel implementations, task metadata, execution logs, and interface-level QA supervision grounded in public ROCm library code. It is not intended to contain personally identifying information or offensive content, and the data construction pipeline filters generated artifacts for executability, determinism, and task validity. Generated kernels are filtered through compilation, execution-based correctness checks, and ROCm validation; QA entries are grounded in retrieved public code contexts rather than user data.

AI assistance.

We used AI assistants during this project for two purposes. First, frontier LLMs such as GPT-5 were used inside HIPKernelGen/TritonKernelGen to generate candidate GPU kernels and reflection feedback, which were then compiled, tested, profiled, and filtered by the authors’ execution-based pipeline. Second, general AI writing assistants were used to help polish wording and organize the manuscript. All experimental design, dataset filtering, validation, analysis, and final claims were checked and approved by the authors.

Appendix B More Related Work

Production Library Code as Supervision.

Benchmark-derived datasets provide diversity but are typically limited to standalone kernels. In contrast, production library code from projects such as rocBLAS and rocSOLVER captures interface-level implementations spanning multiple files and reflecting production-grade engineering practices. These libraries encode years of architecture-specific optimization and correctness engineering. AMDKernelVault incorporates this signal through an interface-aware extraction pipeline that produces kernel-centric supervision grounded in real HIP code.

Positioning Relative to Multi-Turn RL Kernel Agents.

Prior multi-turn systems, including Kevin (Baronio et al., 2026), demonstrate that iterative execution feedback can improve kernel generation. We do not claim that multi-turn feedback alone is new. Our setting differs in two ways. First, AMD HIP/Triton generation starts from a lower-success regime, where early compilation and correctness rewards are sparse. Second, our goal is to train a model that can be deployed inside an agent-training framework, rather than only evaluate an external LLM called by an agent wrapper. In GEAK-style workflows, repeated LLM calls normally handle generation, reflection, and optimization. We instead train one compact policy on execution-feedback trajectories so that the same model can serve these roles locally under ROCm.

On Comparing to Kevin and Devin-Class Agents.

Kevin is trained to emit raw CUDA via torch.utils.cpp_extension.load_inline on NVIDIA hardware, using CUDA-specific primitives such as __shfl_xor_sync and __shared__ memory. Our evaluation targets AMD GPUs through HIP/ROCm and TritonBench-G/ROCmBench under ROCm. A same-protocol comparison would require Kevin to emit AMD-compatible Triton or HIP kernels, which it was not trained for. We therefore treat Kevin as evidence that RL on execution feedback is effective in CUDA settings, while focusing our empirical comparison on models evaluated under the same GEAK OptimAgent-v2 protocol. Devin is a hosted coding agent without a benchmark-accessible interface exposing the fixed generator/reflector/optimizer role assignment used here, so we do not include it as a direct baseline.

System Target Training Multi-turn feedback Performance signal Role internalization
Kevin CUDA / KernelBench Multi-turn RL Yes Correctness + speedup Kernel refinement, not AMD/GEAK roles
AutoTriton Triton SFT + GRPO/RLVR No explicit multi-role loop Rule + execution rewards No generator/reflector/optimizer loop
GEAK AMD Triton No model training Inference-time only Correctness + latency Separate frontier-LLM calls
AMDKernelVault AMD HIP/Triton SFT + execution-aware RL Yes Compile + correctness + speedup One 8B policy for all LLM roles

References: Kevin (Baronio et al., 2026); AutoTriton (Li et al., 2025b); GEAK (Wang et al., 2025a).

Table 6: Positioning relative to closely related kernel-generation systems.

Appendix C Dataset Details

C.1 HIP Kernel Datasets

C.1.1 HIP-CudaAgent Dataset

The HIP-CudaAgent Dataset is derived from CUDA-Agent-Ops-6K using the HIPKernelGen pipeline:

1. PyTorch Task Normalization. The synthesized operator-level PyTorch tasks are converted into function-style references with input construction and correctness-checking logic.

2. HIP Kernel Generation. An LLM-based generator produces HIP implementations for each normalized PyTorch task. Each candidate is compiled with the ROCm toolchain and validated against the PyTorch reference through execution-based tests.

3. Iterative Refinement. Compilation errors, runtime failures, and numerical mismatches trigger reflector-guided regeneration until correctness is achieved or the maximum attempt limit is reached.

Final: 5,388 validated PyTorch-to-HIP samples.

C.1.2 HIP-GPUMode Dataset

The HIP-GPUMode Dataset is generated from the GPUMODE-KernelBook dataset using the HIPKernelGen pipeline:

  1. 1.

    PyTorch Module to Function Conversion. The first-stage agent converts PyTorch module-level code into function-style PyTorch code with correctness-checking logic to filter functionally equivalent pairs.

  2. 2.

    HIP Kernel Generation. The second-stage agent translates each PyTorch module into HIP code using an LLM-based generator. Each candidate undergoes compilation, execution-based correctness validation, and latency measurements.

  3. 3.

    Iterative Refinement. Failed kernels trigger reflector-guided feedback for iterative regeneration until valid or maximum attempts reached.

Final: 5,910 unique PyTorch modules with 22,397 HIP kernel optimization variants (avg 3.8 variants/entry).

C.1.3 HIP-to-HIP Optimization Pairs

We extend HIPKernelGen to support kernel-level HIP-to-HIP optimization by operating directly on existing HIP kernel functions annotated with __global__ or __device__. The input is a baseline HIP kernel and the output is an optimized kernel with equivalent semantics. Final: 34,368 validated HIP kernel optimization pairs.

C.2 Triton Kernel Dataset

We construct a large-scale Triton kernel dataset using TritonKernelGen, aggregating multiple complementary sources under a unified validation pipeline.

C.2.1 Stack-v2-dedup (curated)

We filter and extract PyTorch code containing torch operators and/or Torch Inductor constructs, wrap entries into function-style references, and validate against a high-coverage test suite. From \sim13,000 candidates, we obtain 2,269 medium-difficulty, functionally correct Triton–PyTorch pairs.

C.2.2 TritonBench-train (modified)

We curate 7,713 samples from the original TritonBench-8k. Synthetic kernels generated via the ninetoothed DSL are refined or discarded to remove DSL-specific artifacts and input-specific patterns. We expand test coverage to ensure generalization across shapes, strides, and precision types.

C.2.3 GPUMODE-KernelBook

We incorporate \sim18,000 PyTorch entries with Triton code extracted via Torch Inductor, using them as a primary source for operator coverage and structural guidance.

C.2.4 AI-CUDA-Engineer-Archive

We leverage 11,911 usable PyTorch + CUDA pairs. Where Triton code is absent, the agent synthesizes Triton implementations against the shared PyTorch reference and expanded test suite.

Summary.

Across sources, TritonKernelGen aggregates 39,893 validated samples. Available metadata varies by source subset and packaged representation; where present, it records construction-time validation outcomes, latency, and precision/shape coverage.

C.3 ROCm Libraries QA Dataset

Interface-level extraction from AMD ROCm Libraries using depth-first call graph traversal:

  • rocSOLVER: 519 entries (226 kernel implementations, 293 QA explanations). Difficulty: 190 L1, 198 L2, 131 L3.

  • rocBLAS: 1,858 entries (1,169 implementations, 689 QA explanations). Difficulty: 971 L1, 506 L2, 381 L3.

Total: 2,377 production-grounded supervision entries capturing production HIP development patterns, interface-driven kernel logic, and host–device orchestration.

C.4 Difficulty Distribution

Table 7 shows the full difficulty statistics used to support the curriculum discussion in Section 3. Level 1 contains standalone kernels with relatively direct data flow, Level 2 covers fused or single-file implementations requiring coordinated memory access, and Level 3 includes multi-file or interface-level code with cross-module dependencies. The Triton corpus is dominated by Level 2 examples, while the ROCm Libraries QA subset contains a larger fraction of Level 3 entries because production interfaces often involve host-side orchestration and multi-file kernel dispatch.

Level Triton rocBLAS rocSOLVER
Level 1 (Standalone) 11,824 (29.6%) 971 190
Level 2 (Fused/single-file) 24,765 (62.1%) 506 198
Level 3 (Multi-file) 3,304 (8.3%) 381 131
Table 7: Difficulty distribution across datasets.

Appendix D Dataset Analysis

This section analyzes the resulting datasets in terms of scale, operator coverage, diversity, and architecture-relevant properties.

D.1 Dataset Summary

Table 1 summarizes the scale and composition of AMDKernelVault. The corpus contains 64K+ HIP/ROCm supervision entries and approximately 40K Triton kernels, spanning PyTorch-to-kernel generation, kernel-to-kernel optimization, and production-grounded code understanding. The HIP/ROCm portion combines HIP-CudaAgent and HIP-GPUMode for PyTorch-to-HIP generation, HIP2HIP for optimizing existing HIP kernels, and ROCm Libraries QA for interface-grounded supervision from rocBLAS and rocSOLVER.

The Triton portion aggregates four complementary sources into the Triton-Curated corpus: curated web/code examples from Triton-Stack, benchmark-derived kernels from Triton-Bench, Torch-Inductor-derived programs from Triton-GPUMode, and AI-CUDA-Engineer-derived tasks from Triton-AICE. Together, these subsets provide broad AMD-validated coverage across generated operator workloads, benchmark-style kernels, and production-grounded HIP/Triton programming patterns.

D.2 Operator and Workload Coverage

Across both HIP and Triton datasets, kernels span common deep learning operator families, including elementwise transforms, reductions, normalization layers (e.g., layer normalization), matrix and tensor transformations, and composite fusion patterns produced by Torch Inductor.

Because workloads are derived from real PyTorch programs and benchmark suites, kernels reflect practical memory access patterns, shape variability, and fusion structures encountered in training and inference workloads, rather than synthetic microkernels.

The ROCm Libraries QA dataset extends coverage to production-grade linear algebra primitives from rocBLAS (e.g., GEMM, TRSM, SYMM) and rocSOLVER (e.g., QR factorization, eigensolvers), providing supervision grounded in multi-file, interface-driven kernel implementations.

D.3 Kernel Diversity and Optimization Variants

The datasets exhibit substantial diversity in valid kernel implementations for the same functional operators. In HIP-GPUMode, multiple HIP kernels are provided per PyTorch entry, differing in block sizes, thread mapping strategies, tiling schemes, memory access orderings, and reduction structures.

Similarly, Triton-Curated contains multiple kernel realizations for overlapping operators, capturing variation in program structure, scheduling decisions, and memory layouts at the DSL level. This diversity exposes non-trivial performance trade-offs and supports studies in kernel ranking, selection, and agent-driven optimization.

Such multiplicity of correct implementations is essential for learning-based systems that aim to reason over optimization spaces rather than merely reproducing reference kernels.

D.4 Data Quality and Bias Mitigation

Because a large fraction of the corpus is produced by automated agents, we explicitly filter for execution quality rather than trusting raw model outputs. Candidates must compile, pass reference-based tests, and satisfy ROCm execution constraints before entering the released set. For performance-oriented subsets, multiple candidates are generated and the retained variants are selected after latency measurement on AMD hardware. Public expert-level HIP kernels are scarce, so the corpus combines production kernels from ROCm Libraries with rigorously filtered generated variants. This design removes obviously poor kernels while preserving multiple correct implementations when they expose different optimization strategies.

We also mitigate NVIDIA-centric bias in four ways. First, CUDA-derived samples are treated as seeds, not final AMD supervision: they are translated, compiled, and filtered under ROCm. Second, generated HIP variants are re-optimized and measured on AMD hardware rather than accepted solely for syntactic portability. Third, ROCm Libraries provide expert-written HIP examples from rocBLAS and rocSOLVER, anchoring the dataset in production AMD code. Fourth, Triton samples are refactored and tested for shape generality, stride awareness, and AMD-compatible scheduling assumptions. These checks reduce reliance on NVIDIA-specific warp-size, tensor-core, and tiling assumptions while still preserving useful algorithmic structure from upstream corpora.

D.5 Portability and Correctness Outcomes

All released HIP kernels compile and pass execution-based correctness checks on AMD ROCm after filtering. Triton-Curated kernels likewise compile and pass correctness validation on AMD GPUs under expanded test suites covering diverse shapes, strides, and precision modes.

Because upstream corpora differ substantially in portability, Portability filtering removes kernels that rely on non-general indexing, device-specific assumptions, or input-specialized schedules. As a result, the released datasets represent AMD-valid kernel implementations rather than raw collections from heterogeneous sources.

D.6 Validation Metadata

Validation metadata is subset-dependent. Where available, packaged records include source identifiers, task or reference information, dtype/shape coverage, compiler or runtime status, latency, and speedup relative to the relevant baseline. Some packaged verification fields are empty even when construction-time validation was performed, and rejected-candidate logs were not retained uniformly across pipelines. For ROCm Libraries QA, available provenance can include source paths and interface-level context. We therefore do not claim that every released record contains all fields needed to reconstruct the complete validation trajectory.

D.7 Architecture-Relevant Properties (Triton)

To assess whether Triton-Curated reduces common vendor-shaped artifacts present in upstream corpora, we analyze kernel parameterization and code-shape properties that often encode device-specific assumptions, such as distributions of num_warps, num_stages, and block-size choices.

Compared to raw crawled or DSL-generated Triton code, Triton-Curated exhibits greater diversity in scheduling parameters and fewer hard-coded constants tied to specific hardware configurations. This reflects both execution-based filtering on AMD and agent-driven refactoring toward stride-aware and shape-general indexing, which are required for correctness across ROCm backends. During generation and validation, prompts and feedback expose AMD-specific constraints such as wavefront behavior, LDS capacity, compute-unit occupancy, and target architectures including gfx942 and gfx950.

D.8 Comparison with Existing GPU Kernel Datasets

Existing kernel datasets and benchmarks primarily target CUDA or Nvidia-backed Triton execution, limiting their applicability to ROCm-based platforms. In contrast, HIP-CudaAgent and HIP-GPUMode provide native HIP kernels that execute on AMD GPUs, while Triton-Curated supplies an AMD-validated Triton corpus aligned with the same PyTorch references.

This enables systematic research on AMD-focused kernel optimization, cross-platform transpilation, and heterogeneous GPU performance modeling that is difficult to conduct using Nvidia-centric benchmarks alone.

D.9 Implications for Learning-Based Kernel Research

By combining real-world workload coverage, verified executability on AMD GPUs, and diverse optimization variants across both HIP and Triton representations, the released datasets support supervised fine-tuning, reinforcement learning, and agent-based kernel synthesis under realistic hardware constraints.

These properties enable research on performance-aware kernel generation, automatic optimization strategies, and cross-abstraction learning, substantially lowering the barrier to data-driven GPU systems research on AMD platforms.

Appendix E More Details of Compact LLM Training

This appendix provides implementation details for the training recipe summarized in Section 4.

E.1 Task Serialization and SFT Objective

Each training instance is serialized as an instruction–response pair. The instruction contains the task specification, such as PyTorch reference code, a natural-language Triton prompt, or an input HIP/Triton kernel. The response contains the target kernel implementation, optionally preceded by reasoning annotations for Triton examples.

SFT optimizes the standard next-token likelihood:

SFT(θ)=𝔼(x,y)[logπθ(yx)],\mathcal{L}_{\text{SFT}}(\theta)=-\mathbb{E}_{(x,y^{*})}\left[\log\pi_{\theta}(y^{*}\mid x)\right], (1)

where xx is the task specification and yy^{*} is the target implementation.

For HIP, training examples use direct instruction–response formatting without additional reasoning augmentation. For Triton, when available, we include Chain-of-Thought-style planning before the final kernel:

y=[reasoning,kernel].y^{*}=[\text{reasoning},\text{kernel}].

This helps the model learn tiling choices, masking logic, memory-access patterns, and boundary handling. All SFT runs use full-parameter fine-tuning rather than PEFT/LoRA; in preliminary experiments, parameter-efficient adapters underperformed for this setting because the model has to learn both low-level kernel syntax and agentic self-reflection behavior, and the coupled generation and debugging skills required broader updates than adapter-only training provided.

E.2 Reward Design

Execution-aware RL uses a hierarchical reward:

r=rcompile+rcorrect+rperf,r=r^{\text{compile}}+r^{\text{correct}}+r^{\text{perf}}, (2)

where

rcompile=0.4,rcorrect=1.0,rperf1.5.r^{\text{compile}}=0.4,\quad r^{\text{correct}}=1.0,\quad r^{\text{perf}}\leq 1.5.

The compilation reward is assigned when the candidate builds successfully under the target backend. The correctness reward is assigned when the candidate passes numerical validation against the PyTorch reference. The performance reward is defined as

rperf=min(0.375(log2s)2,1.5),r^{\text{perf}}=\min\left(0.375\cdot(\log_{2}s)^{2},1.5\right), (3)

for speedup s1s\geq 1.

This design gives partial credit to near-valid kernels that compile but fail correctness, which is important in the low-success regime of kernel RL. The bounded performance term rewards meaningful speedups while reducing sensitivity to noisy latency measurements, JIT effects, or autotuning outliers. Typical fast-converging speedups during RL sit well below the 1.5 cap, so this cap functions as a conservative stability guard rather than a tuned hyperparameter; the reward-sensitivity analysis (Appendix J, Table 13) confirms low sensitivity to the exact cap value.

E.3 GRPO Optimization

We use Group Relative Policy Optimization (GRPO) for execution-aware RL (Shao et al., 2024). Standard GRPO normalizes advantages by reward variance:

A^i=rir¯σr+ϵ.\hat{A}_{i}=\frac{r_{i}-\bar{r}}{\sigma_{r}+\epsilon}.

In our setting, many early groups contain uniformly failed samples or samples with very similar partial rewards. We therefore disable variance normalization and use:

A^i=rir¯.\hat{A}_{i}=r_{i}-\bar{r}. (4)

The policy update follows PPO-style clipping:

(θ)=𝔼[min(ρtA^t,clip(ρt,1±ϵ)A^t)]+βDKL.\mathcal{L}(\theta)=-\mathbb{E}\left[\min\left(\rho_{t}\hat{A}_{t},\mathrm{clip}(\rho_{t},1\pm\epsilon)\hat{A}_{t}\right)\right]+\beta D_{\mathrm{KL}}. (5)

E.4 Multi-Turn On-Policy Reflection

For Triton, each rollout follows a multi-turn generate–evaluate–reflect process. At turn tt, the policy observes the task prompt, previous candidate code, compiler output, correctness failures, and optional latency feedback. It then emits both a reflection and a revised kernel. The multi-turn return is:

G(τ)=t=1Tγt1rt,G(\tau)=\sum_{t=1}^{T}\gamma^{t-1}r_{t}, (6)

with γ=0.6\gamma=0.6 and T=3T=3.

Because the same policy produces both reflection and code, the model learns feedback interpretations that are actionable for its own generation behavior. This aligns training with deployment in GEAK-style agent loops, where the trained model serves as generator, reflector, and optimizer.

E.5 Curriculum and Hyperparameters

HIP training uses SFT followed by single-turn GRPO. Triton training uses SFT followed by multi-turn GRPO with on-policy reflection and curriculum scheduling. The Triton curriculum gradually shifts from easier kernels to harder samples, reducing the fraction of Level-1 examples while increasing Level-2 and Level-3 tasks.

We train Qwen3-8B-Instruct with the following configuration:

  • HIP SFT: 8×\timesMI250, 4 epochs.

  • HIP RL: 8×\timesMI325X, 2 epochs, batch size 12, 6 rollouts.

  • Triton SFT: 32×\timesMI325X, 2 epochs.

  • Triton RL: multi-turn GRPO, 2.7 epochs, 3 turns.

E.6 Reward Sensitivity and Design Rationale

Binary correctness rewards are insufficient in early AMD kernel RL because most candidates fail to compile or fail numerical validation. Under such sparse feedback, many GRPO groups produce zero or near-zero useful advantage. The compile reward increases the fraction of samples with usable signal, while disabled variance normalization prevents small reward differences from being amplified or collapsed by unstable group statistics.

Ablations in Appendix J and Table 13 evaluate these choices. Removing multi-turn reflection, removing SFT cold start, or weakening the hierarchical reward reduces performance, confirming that execution-aware training is important for turning AMDKernelVault into an effective training resource for compact AMD kernel LLMs.

Appendix F Evaluation Details

Benchmark scope.

Our execution-intensive evaluation uses 200 PyTorch-to-HIP tasks, 100 HIP-to-HIP optimization tasks, 184 TritonBench-G kernels, and 31 ROCmBench kernels. These sets require compilation and on-device validation and are suitable for controlled comparisons, but they do not establish performance over every AMD operator or workload family.

Agent-budget sensitivity.

Our main results use a fixed 3-iteration budget. On TritonBench-G, Q3-8B+SFT+RL improves from 12.0% Corr@1 to 33.2% Corr@3, compared with 7.6% to 29.8% for Claude Sonnet 4 and 3.8% to 15.2% for GPT-5. On ROCmBench, it improves from 22.58% to 41.94%, compared with Claude Sonnet 4 from 19.35% to 35.48%.

Appendix G Training Details

G.1 Hyperparameters

Parameter HIP Triton
Model
Base model Qwen3-8B-Instruct
Sequence length 8K 16K
SFT
Hardware 8×\timesMI250 32×\timesMI325X
Learning rate 2×1052\times 10^{-5} 1×1051\times 10^{-5}
Batch size 8/GPU 8/GPU
Epochs 4 2
Framework LLaMA-Factory LLaMA-Factory
RL: GRPO / Multi-Turn GRPO
Hardware 8×\timesMI325X 32×\timesMI325X
Learning rate 1×1061\times 10^{-6} 1×1061\times 10^{-6}
Batch size 12 64 prompts
Group size GG 6 8
Clip ϵ\epsilon 0.2 0.2 / 0.28
KL coefficient β\beta 0.01 0.01
Framework VeRL SLIME
Epochs 2 2.7
Multi-Turn (Triton only)
Turns TT 3
Discount γ\gamma 0.6
Reward
Compile partial 0.4
Correct 1.0
Performance cap 1.5
Table 8: Complete hyperparameter configuration.

G.2 Implementation Components

Our implementation separates training, execution feedback, and benchmark scoring into modular components. The multi-turn RL trainer follows a SLIME-style rollout loop: prompts are sampled from the kernel task dataset, the policy generates candidate code and reflections for up to T=3T=3 turns, and GRPO updates the same policy using the hierarchical rewards described in Section 4. Reward-standard-deviation normalization is disabled to preserve stable advantages in the sparse-reward regime. A separate execution server compiles candidates, runs correctness tests, and measures latency, returning compiler/runtime traces and performance feedback to the rollout loop. Final benchmark scoring uses TritonBench-G and ROCmBench, while the headline Triton results use the fixed agent-training evaluator (GEAK OptimAgent-v2) with num_offsprings=1, max_iteration=3, and no profiler feedback.

G.3 Curriculum Learning Phases (Triton)

For Triton RL training, we employ difficulty fading across three phases:

Phase Composition Samples
1 L1 only (Easy) 5.3K
2 L1subset{}_{\text{subset}} + L2 + L3med{}_{\text{med}} 16.5K
3 L2subset{}_{\text{subset}} + L3 + variants 4K
Table 9: Curriculum phases for Triton RL training.

Unlike cumulative curricula that add harder samples while keeping all easy samples, difficulty fading progressively reduces easy samples to focus training on harder cases. This approach prevents the model from overfitting to easy examples while ensuring sufficient exposure to challenging multi-file and fused kernel implementations.

Appendix H Training Dynamics

Figure 2 presents the training dynamics for our multi-turn on-policy reflection RL on Triton kernel generation. The reward trajectory (a) shows steady improvement from the SFT baseline (\sim0.19) to peak values above 1.0, demonstrating effective learning from execution feedback. The benchmark evaluation (b) tracks performance on TritonBench-G (compilation and correctness accuracy) and ROCmBench throughout training, showing consistent gains across both benchmarks.

Figure 2: RL training results for Triton kernel generation. (a) Reward trajectory showing improvement from SFT baseline (\sim0.19) to peak values above 1.0. (b) Benchmark evaluation during training.

Appendix I Additional Experimental Analysis

Cost and deployability.

A practical advantage of training one compact policy is that the GEAK-style loop can run locally on AMD GPUs. Frontier-LLM-based pipelines require repeated external calls for generation, reflection, and optimization, introducing API cost, latency variance, and deployment constraints. In contrast, our trained Qwen3-8B model runs within the local AMD environment and serves all three roles in the agent loop.

Cross-hardware transfer.

To test whether learned optimizations overfit to one AMD target, we evaluate HIP kernels optimized on MI325 GPUs on MI250 GPUs. The optimized kernels retain substantial gains, reaching Speed@10 of 1.68×\times on MI250. Triton evaluations on MI325 and MI350 show similar qualitative trends, suggesting that the model learns transferable optimization principles such as locality, tiling, and parallel reduction patterns while still benefiting from ROCm-specific validation.

Additional failure modes.

Beyond D5 expert-level failures, we observe three common categories of errors: shape-specialized indexing that fails on broader test cases, numerically unstable reductions, and inefficient memory layouts inherited from CUDA/NVIDIA-centric assumptions. These findings motivate future extensions with profiler-guided training feedback and additional production AMD kernels.

Appendix J Ablation Studies

We conduct comprehensive ablation studies to validate our design choices across both HIP and Triton training pipelines.

J.1 Data Scale (HIP)

Table 10 evaluates the impact of training data scale on PyTorch-to-HIP translation (200 tasks from KernelBench Levels 1 and 2).

Data Rate Comp Acc Corr Acc
0% (base) 50.0% 14.5%
10% 62.5% 17.5%
30% 64.5% 18.0%
50% 69.5% 23.0%
100% 83.5% 34.0%
Table 10: Impact of training data scale on PyTorch-to-HIP translation.

With only 10% of training data, the model achieves 62.5% compilation and 17.5% correctness—already surpassing the base model (50.0%/14.5%). Performance scales consistently, confirming our dataset provides high-quality, diverse supervision.

J.2 Triton RL Components

Table 11 presents ablations on the Triton training framework, evaluated using the agent-training evaluator (GEAK OptimAgent-v2) with 3 iterations on TritonBench-G (184 kernels).

Configuration Corr Acc Δ\Delta
Full system 33.2%
Single-turn (T=1T=1) 20.1% -13.1
No SFT cold start 8.5% -24.7
Table 11: Ablation studies on Triton generation.

Key findings:

  • Multi-turn is essential (+13.1%): Single-turn generation cannot recover from initial failures, while 3-turn reflection enables progressive refinement. The model learns to identify errors from execution feedback and generate targeted fixes.

  • SFT cold start is critical (+24.7%): Without SFT initialization, the base model’s low initial success rate (\sim15%) produces extremely sparse RL gradients. SFT raises this to \sim51%, enabling effective RL optimization.

J.3 Fixed Agent Scaffold Ablation

Table 12 isolates the effect of the model training stage under a fixed agent-training evaluator. All rows use the same GEAK OptimAgent-v2 loop on TritonBench-G with num_offsprings=1, max_iteration=3, no profiler feedback, and identical generator/reflector/optimizer role assignment; only the model checkpoint changes. The large gain from base Q3-8B to Q3-8B+SFT+Agentic RL shows that the external agent scaffold alone is insufficient: the policy must be trained on agent feedback trajectories to use compiler, correctness, and performance feedback effectively.

Model inside agent evaluator Training signal Corr@3
Q3-8B (base) None / role-agnostic base model 4.9%
Q3-8B + SFT Supervised kernel data 10.3%
Q3-8B + SFT + Agentic RL Multi-turn compiler/test/perf feedback 33.2%
Table 12: Effect of agentic training under a fixed agent scaffold. All rows use the same GEAK OptimAgent-v2 loop; only the model checkpoint changes.

J.4 Reward Sensitivity

We further analyze the components of the hierarchical reward used during Triton RL. The dominant sensitivity is not the absolute reward scale but whether the reward creates usable gradient signal in the low-success regime.

Factor Sensitivity Observed effect
Compile partial credit High Raises usable gradient mass from near-zero to \sim44%
Disable σ\sigma-normalization Critical Avoids advantage collapse (Adv. 0.5\approx 0.5 vs. 108\sim 10^{-8})
Credit value 0.2 vs. 0.4 Low–medium Both learn; 0.4 gives denser early compile signal
Performance cap Low Stabilizes late training against noisy >>15×\times speedups
Table 13: Qualitative sensitivity of reward design choices in sparse Triton RL.

These observations motivate the chosen reward in Eq. 2: partial compilation credit provides early learning signal, disabled normalization prevents variance collapse or explosion, and the bounded log-squared performance term rewards real speedups without letting measurement jitter dominate correctness.

Appendix K Extended Results

K.1 Pass@K Analysis (Parallel Scaling)

Table 14 presents pass@k results on ROCmBench (31 Triton kernels), measuring the probability of generating at least one correct kernel within kk samples. Our model achieves the highest pass@k across all values, with pass@5 reaching 41.94% compared to 38.71% for Claude Sonnet 4.

Model pass@1 pass@2 pass@3 pass@4 pass@5
Ours 27.10% 35.16% 38.71% 40.65% 41.94%
Claude Sonnet 4 25.81% 33.87% 36.77% 38.06% 38.71%
GPT-5 20.65% 28.39% 31.29% 32.26% 32.26%
Gemini 2.5 Pro 17.42% 24.84% 29.03% 32.26% 35.48%
Table 14: Pass@K results on ROCmBench (31 kernels, 5 samples each with temp=1.0)
Figure 3: Pass@K comparison on ROCmBench (31 kernels). Our model achieves the highest pass rate at all kk values, demonstrating superior sample efficiency for kernel generation.

K.2 Full Difficulty Breakdown

Table 15 provides the complete difficulty-wise breakdown including all model variants, showing the progression from base model through SFT to RL.

Difficulty Claude GPT-5 Gemini Base SFT Ours
D1 (Easy, 3) 1/3 0/3 2/3 2/3 3/3 3/3
D2 (Med, 27) 12/27 7/27 9/27 7/27 10/27 17/27
D3 (Hard, 65) 24/65 16/65 21/65 0/65 6/65 28/65
D4 (V.Hard, 84) 18/84 5/84 11/84 0/84 0/84 13/84
D5 (Expert, 5) 0/5 0/5 0/5 0/5 0/5 0/5
Total 55 28 43 9 19 61
Table 15: Complete difficulty-wise results (exec-correct kernels at final iteration). Base = Qwen3-8B, SFT = Cold-Start SFT, Ours = SFT \rightarrow Multi-Turn RL.

K.3 Training Progression

The progression from Base (4.9%) to SFT (10.3%) to RL (33.2%) demonstrates the effectiveness of each training stage:

  • Base \rightarrow SFT: +5.4% (2.1×\times) — learns kernel syntax and patterns

  • SFT \rightarrow RL: +22.9% (3.2×\times) — learns from execution feedback

Appendix L Theoretical Analysis

L.1 Zero-Advantage Collapse in GRPO

Standard GRPO uses binary rewards (correct/incorrect) with group-normalized advantages. We first prove why this leads to frequent zero-gradient updates, then explain how our design avoids this failure mode.

L.1.1 Binary Reward Collapse (Motivation)

Theorem 1 (Binary Reward Collapse).

In standard GRPO with group size GG and binary rewards r{0,1}r\in\{0,1\}, let pep_{e} denote the probability that a single sample succeeds (receives reward 1). The probability that all samples in a group receive identical rewards, yielding zero advantages, is:

P(A^i=0,i)=peG+(1pe)GP(\hat{A}_{i}=0,\forall i)=p_{e}^{G}+(1-p_{e})^{G} (7)
Proof.

In standard GRPO, the advantage for sample ii is:

A^i=rir¯σr+ϵ\hat{A}_{i}=\frac{r_{i}-\bar{r}}{\sigma_{r}+\epsilon} (8)

where r¯=1Gj=1Grj\bar{r}=\frac{1}{G}\sum_{j=1}^{G}r_{j} is the group mean and σr\sigma_{r} is the group standard deviation.

For binary rewards ri{0,1}r_{i}\in\{0,1\}, the advantage A^i=0\hat{A}_{i}=0 for all ii if and only if all samples receive the same reward:

  • If all ri=0r_{i}=0: then r¯=0\bar{r}=0, so rir¯=0r_{i}-\bar{r}=0 for all ii

  • If all ri=1r_{i}=1: then r¯=1\bar{r}=1, so rir¯=0r_{i}-\bar{r}=0 for all ii

Since samples are drawn independently:

P(all fail)\displaystyle P(\text{all fail}) =(1pe)G\displaystyle=(1-p_{e})^{G} (9)
P(all pass)\displaystyle P(\text{all pass}) =peG\displaystyle=p_{e}^{G} (10)
P(zero advantage)\displaystyle P(\text{zero advantage}) =peG+(1pe)G\displaystyle=p_{e}^{G}+(1-p_{e})^{G} (11)

Corollary 1 (Collapse Rate with Binary Rewards).

For typical kernel generation with \sim6% initial success rate (pe=0.06p_{e}=0.06) and group size G=8G=8:

P(zero advantage)\displaystyle P(\text{zero advantage}) =(0.06)8+(0.94)8\displaystyle=(0.06)^{8}+(0.94)^{8} (12)
0+0.606=60.6%\displaystyle\approx 0+0.606=60.6\% (13)

Over 60% of training batches produce zero gradients—the model cannot learn.

L.1.2 Our Solution: Hierarchical Rewards + Disabled Normalization

To avoid binary-reward collapse, we combine two mechanisms:

1. Hierarchical Rewards.

Instead of binary r{0,1}r\in\{0,1\}, we use:

r=rcompile0.4+rcorrect1.0+rperf1.5[0,2.9]r=\underbrace{r^{\text{compile}}}_{0.4}+\underbrace{r^{\text{correct}}}_{1.0}+\underbrace{r^{\text{perf}}}_{\leq 1.5}\in[0,2.9] (14)

This creates intermediate reward levels: a sample that compiles but fails correctness receives r=0.4r=0.4, distinguishing it from complete failures (r=0r=0). This makes identical rewards across a group less likely.

2. Disabled σ\sigma-Normalization.

Even with hierarchical rewards, early training can exhibit reward clustering (e.g., most samples get r=0r=0 or r=0.4r=0.4). When σr\sigma_{r} is small but non-zero, the standard normalization A^i=(rir¯)/σr\hat{A}_{i}=(r_{i}-\bar{r})/\sigma_{r} amplifies small differences into large, unstable gradients.

During training, we observed training collapse due to these unstable gradients when using standard normalization. Disabling normalization entirely resolved this:

A^i=rir¯\hat{A}_{i}=r_{i}-\bar{r} (15)

The combination ensures: (1) hierarchical rewards create variance by distinguishing partial successes, and (2) disabled normalization keeps gradients stable when that variance is small.

Curriculum-Induced Clustering.

During curriculum training, batches contain samples of similar difficulty (e.g., all L1-Easy). Even successful samples receive similar rewards (\approx1.4), causing σr0\sigma_{r}\approx 0. Disabled normalization handles this gracefully—any differential progress (one sample achieving performance bonus while others don’t) produces non-zero gradients.

L.2 On-Policy Reflection

At each turn, the policy conditions on the prompt, prior code and reflections, and the compiler or execution feedback returned by the environment. The same parameters generate subsequent revisions and are updated from the resulting trajectory return. We use this standard on-policy construction as an implementation choice; we do not claim a new policy-gradient result.

L.3 Multi-Turn Credit Assignment

Lemma 1 (Temporal Credit with γ<1\gamma<1).

For discount factor γ=0.6\gamma=0.6 and T=3T=3 turns, the effective contribution weights are:

w1\displaystyle w_{1} =1.0(Turn 1)\displaystyle=1.0\quad\text{(Turn 1)} (16)
w2\displaystyle w_{2} =0.6(Turn 2)\displaystyle=0.6\quad\text{(Turn 2)} (17)
w3\displaystyle w_{3} =0.36(Turn 3)\displaystyle=0.36\quad\text{(Turn 3)} (18)

After normalization, this assigns approximately 51.0% of the weight to Turn 1, encouraging the model to succeed early while still rewarding improvement through reflection.

Appendix M Multi-Turn Self-Correction Examples

We present abridged pseudocode examples illustrating self-correction on medium-to-hard difficulty kernels. Ellipses and shortened fragments omit boilerplate and therefore are not standalone executable listings.

M.1 Example 1: Cross-Attention (Medium-Hard)

Task: Implement cross-attention softmax kernel for RGB-Depth fusion.

Concepts: Multi-head attention, softmax fusion, batched matrix operations, numerical stability.

Turn 1

(Compile=✓, Exec=✗, Reward=0.4): Missing numerical stability (max subtraction):

@triton.jit
def softmax_kernel(input_ptr, output_ptr, n_rows, n_cols,
BLOCK_SIZE: tl.constexpr):
row_idx = tl.program_id(0)
col_offsets = tl.arange(0, BLOCK_SIZE)
vals = tl.load(input_ptr + row_idx * n_cols + col_offsets)
# BUG: No numerical stability - causes overflow!
output = tl.exp(vals) / tl.sum(tl.exp(vals))
tl.store(output_ptr + row_idx * n_cols + col_offsets, output)

Feedback: “NaN values in output. Softmax requires numerical stability: subtract max before exp() to prevent overflow.”

Turn 2

(Compile=✓, Exec=✓, Reward=1.87, Speedup=2.17×\times): Correct numerically stable softmax:

@triton.jit
def softmax_kernel(input_ptr, output_ptr, n_rows, n_cols,
input_row_stride, BLOCK_SIZE: tl.constexpr):
row_idx = tl.program_id(0)
row_start = input_ptr + row_idx * input_row_stride
col_offsets = tl.arange(0, BLOCK_SIZE)
mask = col_offsets < n_cols
vals = tl.load(row_start + col_offsets, mask=mask, other=-inf)
# FIXED: Numerical stability
row_max = tl.max(vals, axis=0)
vals_shifted = vals - row_max
numerator = tl.exp(vals_shifted)
output = numerator / tl.sum(numerator, axis=0)
tl.store(output_ptr + row_idx * n_cols + col_offsets, output, mask=mask)

Recovery: Immediate (1 turn). Final speedup: 2.17×\times.

M.2 Example 2: PreNorm Transformer Decoder (Hard)

Task: Implement Triton-optimized Pre-Norm Transformer Decoder Layer.

Concepts: Fused layer normalization, multi-head attention, cross-attention, FFN, residual connections.

Turn 1

(Compile=✗, Exec=✗, Reward=0.0): Incorrect kernel structure for layer norm:

@triton.jit
def layer_norm_kernel(x_ptr, weight_ptr, output_ptr, M, N):
# BUG: Missing eps, wrong reduction, no bias
row = tl.program_id(0)
mean = tl.sum(tl.load(x_ptr + row * N)) / N # Wrong load!
var = tl.sum((x - mean) ** 2) / N
output = (x - mean) / tl.sqrt(var) # Missing eps!

Feedback: “Compilation failed: undefined ’x’. Also: missing eps causes division by zero, missing bias parameter.”

Turn 2

(Compile=✓, Exec=✗, Reward=0.4): Fixed syntax but wrong reduction for large N:

@triton.jit
def layer_norm_kernel(x_ptr, weight_ptr, bias_ptr, output_ptr,
M, N, eps: tl.constexpr, BLOCK_SIZE: tl.constexpr):
row = tl.program_id(0)
cols = tl.arange(0, BLOCK_SIZE)
# BUG: Only handles N <= BLOCK_SIZE
x = tl.load(x_ptr + row * N + cols, mask=cols < N)
mean = tl.sum(x) / N
var = tl.sum((x - mean) ** 2) / N
...

Feedback: “Incorrect output for N > BLOCK_SIZE. Use multi-pass reduction for large feature dimensions.”

Turn 3

(Compile=✓, Exec=✓, Reward=2.9, Speedup=9.13×\times): Correct multi-pass layer norm:

@triton.jit
def layer_norm_kernel(x_ptr, weight_ptr, bias_ptr, output_ptr,
M, N, eps: tl.constexpr, BLOCK_SIZE: tl.constexpr):
row = tl.program_id(0)
# FIXED: Multi-pass for arbitrary N
mean = 0.0
for off in range(0, N, BLOCK_SIZE):
cols = off + tl.arange(0, BLOCK_SIZE)
mask = cols < N
x = tl.load(x_ptr + row * N + cols, mask=mask, other=0.0)
mean += tl.sum(x, axis=0)
mean = mean / N
# Similar multi-pass for variance...

Recovery: Progressive (2 turns). Final speedup: 9.13×\times.

M.3 Example 3: Bound Softmax Multi-Pass (Medium)

Task: Implement numerically stable softmax for sequences longer than block size.

Concepts: 3-pass algorithm (max, exp+sum, normalize), arbitrary sequence length handling.

Turn 1

(Compile=✓, Exec=✗, Reward=0.4): Single-pass fails for long sequences:

@triton.jit
def softmax_kernel(input_ptr, output_ptr, n_rows, n_cols,
BLOCK_SIZE: tl.constexpr):
row_idx = tl.program_id(0)
# BUG: Assumes n_cols <= BLOCK_SIZE
cols = tl.arange(0, BLOCK_SIZE)
vals = tl.load(input_ptr + row_idx * n_cols + cols)
max_val = tl.max(vals) # Only partial max!
exp_vals = tl.exp(vals - max_val)
output = exp_vals / tl.sum(exp_vals)

Feedback: “Incorrect softmax for n_cols=2048 > BLOCK_SIZE=1024. Partial max causes incorrect normalization.”

Turn 2

(Compile=✓, Exec=✓, Reward=2.9, Speedup=5.18×\times): Correct 3-pass algorithm:

@triton.jit
def bound_softmax_kernel(input_ptr, output_ptr, n_rows, n_cols,
BLOCK_SIZE: tl.constexpr):
row_idx = tl.program_id(0)
row_start = input_ptr + row_idx * n_cols
# PASS 1: Find global maximum
max_val = float(’-inf’)
for col_off in range(0, n_cols, BLOCK_SIZE):
cols = col_off + tl.arange(0, BLOCK_SIZE)
mask = cols < n_cols
vals = tl.load(row_start + cols, mask=mask, other=float(’-inf’))
max_val = tl.maximum(max_val, tl.max(vals, axis=0))
# PASS 2: Compute exp(x-max) and sum
sum_val = 0.0
for col_off in range(0, n_cols, BLOCK_SIZE):
... # Store intermediate, accumulate sum
# PASS 3: Normalize by sum
for col_off in range(0, n_cols, BLOCK_SIZE):
... # Divide by sum_val

Recovery: Immediate (1 turn). Final speedup: 5.18×\times.

M.4 Summary: Recovery Statistics

Kernel Diff. T1 T2 T3 Speedup
Cross-Attn. Softmax Med.-Hard 2.17×\times
PreNorm Trans. Hard 9.13×\times
Bound Softmax Medium 5.18×\times
Fused Upsample Hard 6.79×\times
Weight Proc. Hard 4.22×\times
Table 16: Multi-turn recovery on medium-to-hard kernels with final speedups. ✓ = compile and execution pass; ✗ = fail.

These examples demonstrate that RL training enables emergent debugging capabilities on complex GPU kernels: the model learns to handle multi-pass reductions, numerical stability, arbitrary tensor dimensions, and fused operations—patterns that require deep understanding of GPU programming paradigms.

M.5 Kernel Quality Analysis

Across the five medium-to-hard case studies summarized above, the agent-trained model produces correct kernels with speedups of 2.17×2.17\times, 9.13×9.13\times, 5.18×5.18\times, 6.79×6.79\times, and 4.22×4.22\times over the PyTorch reference (geometric mean 4.94×\approx 4.94\times, median 5.18×5.18\times). Two qualitative observations emerge from inspecting the generated code. First, the optimizations are not generic templates: the model selects different strategies per workload (multi-pass safe softmax for long sequences, fused layer-norm with running statistics for the Pre-Norm decoder, masked tile loads with ‘tl.exp‘ numerical stabilization for cross-attention). Second, when a turn fails, the next-turn revision typically targets the exact error class in the feedback (e.g., adding numerical stabilization after a NaN report, switching from single-pass to multi-pass reduction after an out-of-bounds error). This pattern is consistent with the multi-turn agent training objective: the same policy must produce both the failure analysis and the corresponding code edit, so reflections are tied to actionable revisions rather than generic suggestions.

We also note what these examples do not show. They are not microbenchmarks tuned for a single shape; each case is validated across multiple shape/stride/dtype combinations under the same test harness used during training, so the reported speedups reflect the realistic correctness-and-performance trade-off our pipeline filters for. A larger-scale latency distribution and memory-consumption study across the full corpus is left for future work and is explicitly listed in our limitations.